pro convert_input_subscans_artemis, dir_out=dir_out, scannum=scannum, array=array, file_scan=file_scan, $
                                    reverse=reverse, signal_scan=signal_scan, debug=debug



;;; conversion of "level-1" xdr files into Scanamorphos input structures, for ArTMiS

;;; to be used one scan at a time: root name must include the scan number
;;; convention Xra = -RA, Ydec = DEC

;;; unused time steps marked by a negative sign in scan.flag_scanleg by ingest_data_artemis


;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; dir_out: directory where the Scanamorphos input structures will be saved
;;; scannum: root name of the input files
;;;          must allow to uniquely select all the files associated with the scan to process
;;;          (i.e. all the legs of the scan)
;;; array: either 350 or 450 (default: 350)

;;; file_scan: ignore (used by format_input_scanam_artemis)


;;; example for N6334 2013 (350 microns):

;;; dir_out='/data/hr/ARTEMIS/n6334_central/'
;;; list_scannum = [43349, 43351, 43353, 43356, 43358, 43359, 43362, 43363, 43365, 43367, 43652]
;;;
;;; for i = 0, n_elements(list_scannum) - 1 do convert_input_subscans_artemis, dir_out=dir_out, $
;;;    scannum=list_scannum(i)

;;; Before running Scanamorphos, include the full path name of dir_out
;;; in the first line of SCANAMORPHOS/scanlist_artemis
;;; and update the list of input structures, one per line.

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



resolve_all, /continue_on_error, /quiet


get_calib_project_info, scannum, init_obs_str, band=array, /silent

pfov = init_obs_str.pfov
rot_angle = -init_obs_str.rot_angle
dx_rcp = init_obs_str.dx_rcp
dy_rcp = init_obs_str.dy_rcp
goodpix_ima = init_obs_str.goodpix_ima

dir_in = init_obs_str.apexdata
nc = strlen(dir_in)
if strmid(dir_in, nc-1) ne '/' then dir_in = dir_in + '/'
obsproject = init_obs_str.project_name
dir_in = dir_in + 'map_otf_xdr/' + obsproject + '/' + strtrim(string(init_obs_str.band), 2) + '/'


nb_line = 16L
nb_col = 18L
off = 3L


scannum = chain(scannum)
array = chain(array)
if (not keyword_set(file_scan)) then file_scan = 'stru_' + scannum + '_' + array + '.xdr'



name_search = dir_in + 'otf_subscan_' + scannum + '*.xdr'
files = file_search(name_search, /fold_case, /test_read, count=nf)
if nf ge 1 then begin
   print, ''
   print, chain(nf) + ' subscans found'
   print, ''


   ;;; check if subscans are time-ordered:

   tstart = fltarr(nf)
   for i = 0L, nf - 1L do begin
      restore, files(i)
      tstart(i) = donnees_red.datapar(0L).lst
   endfor
   tstart_off = tstart - tstart(0L)
   ind = sort(tstart_off)
   files = files(ind)
   print, 'starting time: ' + chain(tstart(ind(0L)), 3)


   ;;; define usage mask to reinject processed data into original cube format at the end:
   restore, files(0)
   npix_x = n_elements(donnees_red.cube(*, 0, 0))
   npix_y = n_elements(donnees_red.cube(0, *, 0))
   if ((npix_x ne 35 and npix_x ne 73) or npix_y ne 39) then begin
      print, 'cube dimensions not accounted for'
      print, 'Check the format of the input data.'
      return
   endif

   mask_use = intarr(npix_x, npix_y)
   narr_x = npix_x / nb_line
   narr_y = npix_y / nb_col
   nb = narr_x * narr_y * nb_line * nb_col
   for nmy = 0, narr_y - 1 do for nmx = 0, narr_x - 1 do $
      mask_use(nmx*(nb_line+off):nmx*(nb_line+off)+nb_line-1L, nmy*(nb_col+off):nmy*(nb_col+off)+nb_col-1L) = 1


   ;;; read distortion file to get correct offsets of each matrix from array center:
   dx_rcp = reform(dx_rcp(where(mask_use gt 0.5)), narr_x*nb_line, narr_y*nb_col)
   dy_rcp = reform(dy_rcp(where(mask_use gt 0.5)), narr_x*nb_line, narr_y*nb_col)

   dx_as_matrix = pfov * rebin(dindgen(narr_x*nb_line), narr_x*nb_line, narr_y*nb_col)
   dy_as_matrix = pfov * dindgen(narr_y*nb_col) ## (1.D0 + dblarr(narr_x*nb_line))
   for nmy = 0, narr_y - 1 do for nmx = 0, narr_x - 1 do begin
      dx_as_matrix(nmx*nb_line:(nmx+1)*nb_line-1L, nmy*nb_col:(nmy+1)*nb_col-1L) += $
         median(dx_rcp(nmx*nb_line:(nmx+1)*nb_line-1L, nmy*nb_col:(nmy+1)*nb_col-1L) $
         - dx_as_matrix(nmx*nb_line:(nmx+1)*nb_line-1L, nmy*nb_col:(nmy+1)*nb_col-1L), /even)
      dy_as_matrix(nmx*nb_line:(nmx+1)*nb_line-1L, nmy*nb_col:(nmy+1)*nb_col-1L) += $
         median(dy_rcp(nmx*nb_line:(nmx+1)*nb_line-1L, nmy*nb_col:(nmy+1)*nb_col-1L) $
         - dy_as_matrix(nmx*nb_line:(nmx+1)*nb_line-1L, nmy*nb_col:(nmy+1)*nb_col-1L), /even)
   endfor

   goodpix_ima = reform(goodpix_ima(where(mask_use gt 0.5)), narr_x*nb_line, narr_y*nb_col)
   w = where(goodpix_ima lt 0.5)
   dx_as_matrix(w) = !values.f_nan
   dy_as_matrix(w) = !values.f_nan
   dx_as_matrix = reform(dx_as_matrix, nb)
   dy_as_matrix = reform(dy_as_matrix, nb)



   ;;; reinject processed data into original structures:
   if keyword_set(reverse) then begin
      if keyword_set(debug) then begin
         print, 'The "donnees_red" structures in the following files will be overwritten:'
         print, files
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      endif
      restore, dir_out + file_scan

      nb = n_elements(signal_scan(0, *))
      nt = n_elements(signal_scan(*, 0))
      if nb gt npix_x * npix_y then begin
         signal_scan = reform(signal_scan, nt, 4L*nb_line, 2L*nb_col)
         signal_scan = signal_scan(*, nb_line:3L*nb_line-1L, *)
         signal_scan = reform(signal_scan, nt, 4L*nb_line*nb_col)
      endif

      for i = 0L, nf - 1L do begin
         restore, files(i)
         nt = n_elements(donnees_red.datapar.lst)
         mask_use_scan = rebin(mask_use, npix_x, npix_y, nt)
         wsubs = where(abs(scan.flag_scanleg) eq i + 1L)
         wz = where(scan.flag_scanleg(wsubs) lt 0, cz, complement=wdef)
         if cz ge 1 then mask_use_scan(*, *, wz) = 0
         w_inject = where(mask_use_scan gt 0.5)

         donnees_red.cube(*, *, *) = !values.f_nan
         donnees_red.cube(w_inject) = transpose(signal_scan(wsubs(wdef), *))
         save, filename=files(i), subscanname, carte_scan, donnees_cal, donnees_red
      endfor
   endif else begin


   ;;; read data:

   for i = 0L, nf - 1L do begin
      restore, files(i)
      donnees = donnees_cal
      if i eq 0L then target = donnees.source

      time_vect = donnees_red.datapar.lst
      nt = n_elements(time_vect)
      if i eq 0L then begin
         time_s = time_vect
         flag_scanleg = i + 1L + lonarr(nt)
      endif else begin
         time_s = [time_s, time_vect]
         flag_scanleg = [flag_scanleg, i + 1L + lonarr(nt)]
      endelse

      if i eq nf - 1L then begin
         vscan_status = donnees_red.scanxvel
         nt_tot = n_elements(time_s)
         dt_status = median(time_s(1L:nt_tot-1L) - time_s(0L:nt_tot-2L))
      endif


      ;;; restructure signal cube:
      if narr_x eq 2 then $
         signal = [donnees.cube(0L:nb_line-1L, *, *), donnees.cube(nb_line+off:2L*nb_line+off-1L, *, *)] $
       else signal = [donnees.cube(0L:nb_line-1L, *, *), donnees.cube(nb_line+off:2L*nb_line+off-1L, *, *), $
         donnees.cube(2L*(nb_line+off):3L*nb_line+2L*off-1L, *, *), donnees.cube(3L*(nb_line+off):4L*nb_line+3L*off-1L, *, *)]
      signal = [[signal(*, 0L:nb_col-1L, *)], [signal(*, nb_col+off:2L*nb_col+off-1L, *)]]
      

      mat_boxes = 1 + intarr(narr_x*nb_line, narr_y*nb_col)
      for arr = 1, narr_x - 1 do mat_boxes(arr*nb_line:(arr+1)*nb_line-1L, *) += arr
      for arr = 1, narr_y - 1 do mat_boxes(*, arr*nb_col:(arr+1)*nb_col-1L) += (arr * narr_x)

      signal = transpose(reform(signal, nb, nt))
      mat_boxes = reform(mat_boxes, nb)


      if i eq 0L then signal_psw = signal else signal_psw = [signal_psw, signal]



      ;;; read sky coordinates of array center:
      subscan_ra_deg = donnees_red.datapar.ra
      subscan_dec_deg = donnees_red.datapar.dec


      ;;; map coordinate vectors of array center in degrees (with Xra = -RA, Ydec = DEC) :
      xra_tan = -subscan_ra_deg
      ydec_tan = subscan_dec_deg
      refx = median(xra_tan)
      refy = median(ydec_tan)
      gnomonic, xra_tan, ydec_tan, refx=refx, refy=refy


      ;;; array orientation in (N, E) frame varies with time:
      pa_array = rot_angle + donnees_red.datapar.parangle
      cos_pa_array = cos(pa_array * !dpi / 180.D0)
      sin_pa_array = sin(pa_array * !dpi / 180.D0)


      dXra_temp = dx_as_matrix ## (1.D0 + dblarr(nt))
      dYdec_temp = dy_as_matrix ## (1.D0 + dblarr(nt))
      cos_pa_array_rebin = rebin(cos_pa_array, nt, nb)
      sin_pa_array_rebin = rebin(sin_pa_array, nt, nb)
      dXra = -dXra_temp * cos_pa_array_rebin - dYdec_temp * sin_pa_array_rebin
      dYdec = dYdec_temp * cos_pa_array_rebin - dXra_temp * sin_pa_array_rebin


      ;;; bolometer map coordinate arrays in degrees:
      Xra_bolos = rebin(xra_tan, nt, nb) + dXra / 3600.D0
      Ydec_bolos = rebin(ydec_tan, nt, nb) + dYdec / 3600.D0

      ;;; apply pointing offsets:
      tags = tag_names(donnees_cal)
      pos = stregex(tags, 'dra_point', /fold_case)
      w = where(pos ge 0, cw)
      if cw ge 1 then begin
         Xra_bolos += donnees_cal.dra_point / 3600.D0
         Ydec_bolos -= donnees_cal.ddec_point / 3600.D0
      endif


      ;;; return to sky coordinates:
      gnomonic_invert, Xra_bolos, Ydec_bolos, refx=refx, refy=refy


      if i eq 0L then begin
         Xra_bolos_psw = Xra_bolos
         Ydec_bolos_psw = Ydec_bolos
      endif else begin
         Xra_bolos_psw = [Xra_bolos_psw, Xra_bolos]
         Ydec_bolos_psw = [Ydec_bolos_psw, Ydec_bolos]
      endelse


      mask = reform(goodpix_ima, nb) ## (1. + fltarr(nt))
      if i eq 0L then mask_psw = mask else mask_psw = [mask_psw, mask]
   endfor

   nt = n_elements(time_s)


   ;;; eliminate invalid data:
   w = where(finite(signal_psw) eq 0 or finite(Xra_bolos_psw) eq 0 or finite(Ydec_bolos_psw) eq 0 $
      or mask_psw lt 1, cw)

   if cw ge 1 then begin
      signal_psw(w) = !values.f_nan
      wb = w / nt
      histo = histogram(wb, min=0., max=nb-1.)
      wkeep = where(histo lt 9L * nt / 10L, complement=wz, ncomplement=nz)
      if nz ge 1 then begin
         print, 'disabling ' + chain(nz) + ' bolometer(s) valid < 10% of the time'
         signal_psw(*, wz) = !values.f_nan
         mat_boxes(wz) = !values.f_nan
         print, ''
      endif
   endif



   toto = reform(signal_psw, nt, narr_x*nb_line, narr_y*nb_col)
   for nmx = 0, narr_x - 1 do for nmy = 0, narr_y - 1 do $
      toto(*, nmx*nb_line:(nmx+1)*nb_line-1L, nmy*nb_col:(nmy+1)*nb_col-1L) -= $
         median(toto(*, nmx*nb_line:(nmx+1)*nb_line-1L, nmy*nb_col:(nmy+1)*nb_col-1L))


   if narr_x eq 2 then begin
      signal_psw = !values.f_nan + fltarr(nt, 4L*nb_line, 2L*nb_col)
      signal_psw(*, nb_line:3L*nb_line-1L, *) = toto
      signal_psw = reform(signal_psw, nt, 8L*nb_line*nb_col)
      toto = reform(Xra_bolos_psw, nt, narr_x*nb_line, narr_y*nb_col)
      Xra_bolos_psw = !values.f_nan + fltarr(nt, 4L*nb_line, 2L*nb_col)
      Xra_bolos_psw(*, nb_line:3L*nb_line-1L, *) = toto
      Xra_bolos_psw = reform(Xra_bolos_psw, nt, 8L*nb_line*nb_col)
      toto = reform(Ydec_bolos_psw, nt, narr_x*nb_line, narr_y*nb_col)
      Ydec_bolos_psw = !values.f_nan + fltarr(nt, 4L*nb_line, 2L*nb_col)
      Ydec_bolos_psw(*, nb_line:3L*nb_line-1L, *) = toto
      Ydec_bolos_psw = reform(Ydec_bolos_psw, nt, 8L*nb_line*nb_col)
      toto = reform(mat_boxes, narr_x*nb_line, narr_y*nb_col)
      mat_boxes = !values.f_nan + intarr(4L*nb_line, 2L*nb_col)
      mat_boxes(nb_line:3L*nb_line-1L, *) = toto
      mat_boxes = reform(mat_boxes, 8L*nb_line*nb_col)
   endif else signal_psw = reform(toto, nt, nb)



   ;;; create structure:
   if array eq 350 then begin
      tag_signal = 'signal_psw'
      tag_Xra = 'Xra_bolos_psw'
      tag_Ydec = 'Ydec_bolos_psw'
   endif else begin
      tag_signal = 'signal_plw'
      tag_Xra = 'Xra_bolos_plw'
      tag_Ydec = 'Ydec_bolos_plw'
   endelse
   scan = create_struct('time_s', time_s, 'flag_scanleg', flag_scanleg, $
      tag_signal, signal_psw, 'mat_boxes', mat_boxes, $
      tag_Xra, Xra_bolos_psw, tag_Ydec, Ydec_bolos_psw, $
      'vscan_status', vscan_status, 'dt_status', dt_status, 'pfov', pfov, $
      'target', target, 'obsproject', obsproject, $
      'units', ['times in seconds', 'fluxes in Jy/beam', 'coordinates in degrees', 'scan speed in arcsec/s'])

   save, filename=dir_out+file_scan, scan, /xdr
   print, 'saved file "' + file_scan + '"'
   print, ''

   endelse

endif else begin
   print, ''
   print, 'no subscan found'
   print, 'Check directory and file names and permissions.'
endelse


end
